@@ -4,7 +4,7 @@ use warnings;
 use 5.010; # state
 
 # ABSTRACT: Fully UTF-8 aware File::Find
-our $VERSION = '0.005'; # VERSION
+our $VERSION = '0.006'; # VERSION
 
 #pod =for test_synopsis
 #pod my @directories_to_search;
@@ -36,7 +36,7 @@ our $VERSION = '0.005'; # VERSION
 #pod This module replaces the L<File::Find> functions with fully UTF-8
 #pod aware versions, both expecting and returning characters.
 #pod
-#pod B<Note:> Replacement of functions is not done on DOS and OS/2
+#pod B<Note:> Replacement of functions is not done on DOS, Windows, and OS/2
 #pod as these systems do not have full UTF-8 file system support.
 #pod
 #pod =head2 Behaviour
@@ -71,37 +71,69 @@ our $VERSION = '0.005'; # VERSION
 #pod
 #pod =cut
 
-use File::Find qw();
-use Encode;
-
-my @EXPORT = qw(find finddepth);
-my @EXPORT_OK = ();
+use File::Find ();
+use Encode ();
 
 # Holds the pointers to the original version of redefined functions
 state %_orig_functions;
 
-# Current package
+# Current (i.e., this) package
 my $current_package = __PACKAGE__;
 
 # Original package (i.e., the one for which this module is replacing the functions)
 my $original_package = $current_package;
 $original_package =~ s/::utf8$//;
 
+require Carp;
+$Carp::Internal{$current_package}++; # To get warnings reported at correct caller level
+
+#pod =attr $File::Find::utf8::SPECIALVARS
+#pod
+#pod By default C<File::Find::utf8> only decodes the I<normal>
+#pod L<File::Find> variables C<$_>, C<$File::Find::name>,
+#pod C<$File::Find::dir>, and (when C<follow> or C<follow_fast> is in
+#pod effect) C<$File::Find::fullname> for use in the C<preprocess>,
+#pod C<wanted>, and C<postporcess> functions.
+#pod
+#pod If for any reason (e.g., compatibility with find.pl or find2perl) you
+#pod also need the I<special> variables C<$File::Find::topdir>,
+#pod C<$File::Find::topdev>, C<$File::Find::topino>,
+#pod C<$File::Find::topmode>, and C<$File::Find::topnlink> to be decoded,
+#pod specify C<local $File::Find::utf8::COMPATILBILITY = 1;> in your
+#pod code. The extra decoding that needs to happen will impact performance
+#pod though, so use only when absolutely necessary.
+#pod
+#pod =cut
+
+our $SPECIALVARS = 0;
+
+#pod =attr $File::Find::utf8::UTF8_CHECK
+#pod
+#pod By default C<File::Find::utf8> marks decoding errors as fatal (default value
+#pod for this setting is C<Encode::FB_CROAK>). If you want, you can change this by
+#pod setting C<File::Find::utf8::UTF8_CHECK>. The value C<Encode::FB_WARN> reports
+#pod the encoding errors as warnings, and C<Encode::FB_DEFAULT> will completely
+#pod ignore them. Please see L<Encode> for details. Note: C<Encode::LEAVE_SRC> is
+#pod I<always> enforced.
+#pod =cut
+
+our $UTF8_CHECK = Encode::FB_CROAK; # Die on encoding errors
+
+# UTF-8 Encoding object
+my $_UTF8 = Encode::find_encoding('UTF-8');
+
 sub import {
     # Target package (i.e., the one loading this module)
     my $target_package = caller;
 
-    no strict qw(refs); ## no critic (TestingAndDebugging::ProhibitNoStrict)
-    no warnings qw(redefine);
-
     # If run on the dos/os2/windows platform, ignore overriding functions silently.
-    # These platforms do have (proper) utf-8 file system suppport...
+    # These platforms do not have (proper) utf-8 file system suppport...
     unless ($^O =~ /MSWin32|cygwin|dos|os2/) {
         no strict qw(refs); ## no critic (TestingAndDebugging::ProhibitNoStrict)
         no warnings qw(redefine);
 
         # Redefine each of the functions to their UTF-8 equivalent
-        for my $f (@EXPORT, @EXPORT_OK) {
+        for my $f (@{$original_package . '::EXPORT'}, @{$original_package . '::EXPORT_OK'}) {
             # If we already have the _orig_function, we have redefined the function
             # in an earlier load of this module, so we need not do it again
             unless ($_orig_functions{$f}) {
@@ -112,37 +144,20 @@ sub import {
         $^H{$current_package} = 1; # Set compiler hint that we should use the utf-8 version
     }
 
-    if ($#_) {
-        # Check arguments
-        my @invalid_exports;
-        for my $f (@_[1..$#_]) {
-            if (! grep /^$f$/, (':none', @EXPORT, @EXPORT_OK)) {
-                push @invalid_exports, "$f is not exported by $current_package module";
-            }
-        }
-        if (@invalid_exports) {
-            require Carp;
-            Carp::croak(join("\n", @invalid_exports)  . "\nCan't continue after import errors");
-        }
-    }
+    # Determine symbols to export
+    shift; # First argument contains the package (that's us)
+    @_ = (':DEFAULT') if !@_; # If nothing provided, use default
+    @_ = map { $_ eq ':none' ? () : $_ } @_; # Replace :none tag with empty list
 
-    # Export functions to target package
-    unless ($#_ && grep /^:none$/, @_[1..$#_]) {
-        for my $f ($#_ ? @_[1..$#_] : @EXPORT) {
-            *{$target_package . '::' . $f} = \&{$original_package . '::' . $f};
-        }
-    }
+    # Use exporter to export
+    require Exporter;
+    Exporter::export_to_level($original_package, 1, $target_package, @_) if (@_);
 
     return;
 }
 
-sub unimport {
-    # If run on the dos/os2/windows platform, ignore overriding functions silently.
-    # These platforms do have (proper) utf-8 file system suppport...
-    unless ($^O =~ /MSWin32|cygwin|dos|os2/) {
-        $^H{$current_package} = 0; # Set compiler hint that we should not use the utf-8 version
-    }
-
+sub unimport { ## no critic (Subroutines::ProhibitBuiltinHomonyms)
+    $^H{$current_package} = 0; # Set compiler hint that we should not use the utf-8 version
     return;
 }
 
@@ -151,47 +166,57 @@ sub _utf8_find {
     #  Make argument always into the find's options hash
     my %find_options_hash = ref($ref) eq "HASH" ? %$ref : (wanted => $ref);
 
-    # Save original processors
-    my %org_proc;
-    for my $proc ("wanted", "preprocess", "postprocess") { $org_proc{$proc} = $find_options_hash{$proc}; }
-
+    # Holds the (possibly encoded) arguments
     my @args = @_;
 
+    # Enforce LEAVE_SRC
+    $UTF8_CHECK |= Encode::LEAVE_SRC if $UTF8_CHECK;
+
     # Get the hint from the caller (one level deeper if called from finddepth)
     my $hints = ((caller 1)[3]//'') ne 'File::Find::utf8::_utf8_finddepth' ? (caller 0)[10] : (caller 1)[10];
     if ($hints->{$current_package}) {
+        # Save original processors
+        my %org_proc;
+        for my $proc ("wanted", "preprocess", "postprocess") { $org_proc{$proc} = $find_options_hash{$proc}; }
+        my $follow_option = (exists $find_options_hash{follow} && $find_options_hash{follow})
+            || (exists $find_options_hash{follow_fast} && $find_options_hash{follow_fast});
+
         # Wrap processors to become utf8-aware
         for my $proc ("wanted", "preprocess", "postprocess") {
             if (defined $org_proc{$proc} && ref $org_proc{$proc}) {
                 $find_options_hash{$proc} = sub {
                     # Decode the file variables so they become characters
-                    local $_                    = decode('UTF-8', $_);
-                    local $File::Find::name     = decode('UTF-8', $File::Find::name);
-                    local $File::Find::dir      = decode('UTF-8', $File::Find::dir);
-                    local $File::Find::fullname = decode('UTF-8', $File::Find::fullname);
-                    local $File::Find::topdir   = decode('UTF-8', $File::Find::topdir);
-                    local $File::Find::topdev   = decode('UTF-8', $File::Find::topdev);
-                    local $File::Find::topino   = decode('UTF-8', $File::Find::topino);
-                    local $File::Find::topmode  = decode('UTF-8', $File::Find::topmode);
-                    local $File::Find::topnlink = decode('UTF-8', $File::Find::topnlink);
-                    $org_proc{$proc}->(@_);
+                    local $_                    = $_UTF8->decode($_,                    $UTF8_CHECK) if $_;
+                    local $File::Find::name     = $_UTF8->decode($File::Find::name,     $UTF8_CHECK) if $File::Find::name;
+                    local $File::Find::dir      = $_UTF8->decode($File::Find::dir,      $UTF8_CHECK) if $File::Find::dir;
+                    local $File::Find::fullname = $_UTF8->decode($File::Find::fullname, $UTF8_CHECK) if $follow_option && $File::Find::fullname;
+                    # These are only necessary for compatibility reasons (find.pl, find2perl).
+                    # If you need them, set $File::Find::utf8::SPECIALVARS
+                    local $File::Find::topdir   = $_UTF8->decode($File::Find::topdir,   $UTF8_CHECK) if $SPECIALVARS && $File::Find::topdir;
+                    local $File::Find::topdev   = $_UTF8->decode($File::Find::topdev,   $UTF8_CHECK) if $SPECIALVARS && $File::Find::topdev;
+                    local $File::Find::topino   = $_UTF8->decode($File::Find::topino,   $UTF8_CHECK) if $SPECIALVARS && $File::Find::topino;
+                    local $File::Find::topmode  = $_UTF8->decode($File::Find::topmode,  $UTF8_CHECK) if $SPECIALVARS && $File::Find::topmode;
+                    local $File::Find::topnlink = $_UTF8->decode($File::Find::topnlink, $UTF8_CHECK) if $SPECIALVARS && $File::Find::topnlink;
+                    return $org_proc{$proc}->(@_);
                 };
             }
         }
         # Encode arguments as utf-8 so that the original File::Find receives bytes
-        @args = map { encode('UTF-8', $_) } @_;
+        @args = map { $_ ? $_UTF8->encode($_, $UTF8_CHECK) : $_ } @_;
     }
 
-    # Make sure warning level propagates
+    # Make sure warning level propagates to File::Find
+    # Note: on perl prior to v5.12 warnings_fatal_enabled does not exist
+    #       so we can not use it.
     if (!warnings::enabled('File::Find')) {
         no warnings 'File::Find';
         return $_orig_functions{find}->(\%find_options_hash, @args);
     } elsif (!exists &warnings::fatal_enabled or !warnings::fatal_enabled('File::Find')) {
         use warnings 'File::Find';
-        return $_orig_functions{find}->(\%find_options_hash, map { encode('UTF-8', $_) } @_);
+        return $_orig_functions{find}->(\%find_options_hash, @args);
     } else {
         use warnings FATAL => qw(File::Find);
-        return $_orig_functions{find}->(\%find_options_hash, map { encode('UTF-8', $_) } @_);
+        return $_orig_functions{find}->(\%find_options_hash, @args);
     }
 }
 
@@ -214,7 +239,7 @@ File::Find::utf8 - Fully UTF-8 aware File::Find
 
 =head1 VERSION
 
-version 0.005
+version 0.006
 
 =head1 SYNOPSIS
 
@@ -243,7 +268,7 @@ characters.
 This module replaces the L<File::Find> functions with fully UTF-8
 aware versions, both expecting and returning characters.
 
-B<Note:> Replacement of functions is not done on DOS and OS/2
+B<Note:> Replacement of functions is not done on DOS, Windows, and OS/2
 as these systems do not have full UTF-8 file system support.
 
 =head2 Behaviour
@@ -261,6 +286,33 @@ L<File::Find> warning levels are properly propagated. Note though that
 for propagation of fatal L<File::Find> warnings, Perl 5.12 or higher
 is required (or the appropriate version of L<warnings>).
 
+=head1 ATTRIBUTES
+
+=head2 $File::Find::utf8::SPECIALVARS
+
+By default C<File::Find::utf8> only decodes the I<normal>
+L<File::Find> variables C<$_>, C<$File::Find::name>,
+C<$File::Find::dir>, and (when C<follow> or C<follow_fast> is in
+effect) C<$File::Find::fullname> for use in the C<preprocess>,
+C<wanted>, and C<postporcess> functions.
+
+If for any reason (e.g., compatibility with find.pl or find2perl) you
+also need the I<special> variables C<$File::Find::topdir>,
+C<$File::Find::topdev>, C<$File::Find::topino>,
+C<$File::Find::topmode>, and C<$File::Find::topnlink> to be decoded,
+specify C<local $File::Find::utf8::COMPATILBILITY = 1;> in your
+code. The extra decoding that needs to happen will impact performance
+though, so use only when absolutely necessary.
+
+=head2 $File::Find::utf8::UTF8_CHECK
+
+By default C<File::Find::utf8> marks decoding errors as fatal (default value
+for this setting is C<Encode::FB_CROAK>). If you want, you can change this by
+setting C<File::Find::utf8::UTF8_CHECK>. The value C<Encode::FB_WARN> reports
+the encoding errors as warnings, and C<Encode::FB_DEFAULT> will completely
+ignore them. Please see L<Encode> for details. Note: C<Encode::LEAVE_SRC> is
+I<always> enforced.
+
 =for test_synopsis my @directories_to_search;
 
 =head1 BUGS
